package cz.cuni.mff.nprg013.tests.robots.model;

import java.awt.Color;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

public class RobotModel {
    private static final Random rnd = new Random();
    public static final long TIME_OUT = 60 * 1000; // 1 MIN
    
    private static final Color[] colors = { Color.WHITE, Color.LIGHT_GRAY, Color.GRAY, Color.DARK_GRAY, Color.RED, Color.PINK, Color.ORANGE, Color.YELLOW, Color.GREEN, Color.MAGENTA, Color.CYAN, Color.BLUE }; 
            
    private final String id;
    private final String name;

    private RobotState state;
    private int stepsExecuted;
    
    private int posX;
    private int posY;
    private RobotOrientation direction;
    
    private Color color;
    private List<Actions> steps; // Executed steps as well as the planned steps
    
    private long lastActivity; // System time
    
    public RobotModel(String id, String name, Maze maze, List<RobotModel> otherRobots) {
        this.id = id;
        this.name = name;
        
        this.state = RobotState.NEED_KILL;
        this.stepsExecuted = 0;
        
        this.direction = RobotOrientation.values()[rnd.nextInt(RobotOrientation.values().length)];
        this.color = colors[rnd.nextInt(colors.length)];
        
        this.steps = new ArrayList<>();
        setRandomPosition(maze, otherRobots);
        
        lastActivity = System.currentTimeMillis();
    }

    // Empty maze (no robots or fires)
    private void setRandomPosition(Maze maze, List<RobotModel> otherRobots) {
        Maze m = maze.addRobots(otherRobots);
        while(true) {
            posX = rnd.nextInt(Maze.WIDTH);
            posY = rnd.nextInt(Maze.HEIGHT);
            
            if (MazeFields.EMPTY == m.getMaze(posY, posX)) {
                break;
            }
        }
    }
    
    public String getId() {
        return id;
    }

    public String getName() {
        return name;
    }

    public int getPosX() {
        return posX;
    }
    
    public int getPosY() {
        return posY;
    }
    
    public RobotState getState() {
        return state;
    }
    
    public Color getColor() {
        return color;
    }
    
    public void performActionIfMovement(Maze m, List<RobotModel> robots) {
        if (!isActive()) {
            return;
        }
        // Is there action to perform
        if (steps.size() <= stepsExecuted) {
            return;
        }
        
       lastActivity = System.currentTimeMillis();
       m = m.addRobots(robots);
        
        Actions action = steps.get(stepsExecuted);
        if (action == Actions.FIRE) {
            return;
        }
        
        if (action == Actions.TURN_RIGHT) {
            this.direction = RobotOrientation.values()[(this.direction.ordinal() + 1) % RobotOrientation.values().length]; 
        } else if (action == Actions.TURN_LEFT) {
            this.direction = RobotOrientation.values()[( RobotOrientation.values().length + this.direction.ordinal() - 1) % RobotOrientation.values().length]; 
        } else if (action == Actions.MOVE_FORWARD) {
            if (direction == RobotOrientation.UP) {
                MazeFields mf = m.getMaze(posY-1, posX);
                if (mf == MazeFields.EMPTY || mf == MazeFields.EXIT) {
                    posY--;
                }
            } else if (direction == RobotOrientation.DOWN) {
                MazeFields mf = m.getMaze(posY+1, posX);
                if (mf == MazeFields.EMPTY || mf == MazeFields.EXIT) {
                    posY++;
                }
            } else if (direction == RobotOrientation.LEFT) {
                MazeFields mf = m.getMaze(posY, posX-1);
                if (mf == MazeFields.EMPTY || mf == MazeFields.EXIT) {
                    posX--;
                }
            } else if (direction == RobotOrientation.RIGHT) {
                MazeFields mf = m.getMaze(posY, posX+1);
                if (mf == MazeFields.EMPTY || mf == MazeFields.EXIT) {
                    posX++;
                }
           }
        }
    }
    
    
    private boolean isActive() {
        return state == RobotState.NEED_KILL || state == RobotState.LOOKING_EXIT;
        
    }

    public void performActionIfFire(Maze m, List<RobotModel> robots) {
        if (!isActive()) {
            return;
        }
        Maze m1 = m.addRobots(robots);
        // Is there action to perform
        if (steps.size() <= stepsExecuted) {
            return;
        }
        
        lastActivity = System.currentTimeMillis();
        
        Actions action = steps.get(stepsExecuted);
        if (action != Actions.FIRE) {
            return;
        }
        
        for(int i = 1; i < Math.max(Maze.WIDTH, Maze.HEIGHT); i++) {
            int dx = 0;
            int dy = 0;
            
            if (direction == RobotOrientation.UP) {
                dy = -1;
            } else if (direction == RobotOrientation.DOWN) {
                dy = 1;
            } else if (direction == RobotOrientation.LEFT) {
                dx = -1;
            } else if (direction == RobotOrientation.RIGHT) {
                dx = +1;
           }
           int px = posX + i * dx;
           int py = posY + i * dy;
           px = Math.min(Math.max(px,  0), Maze.WIDTH);
           py = Math.min(Math.max(py,  0), Maze.HEIGHT);
           
           MazeFields mf = m1.getMaze(py, px);
           if (mf == MazeFields.ROBOT) {
               RobotModel rm = findRobotAtPosition(px, py, robots);
               rm.kill();
               if (state == RobotState.NEED_KILL) {
                   state = RobotState.LOOKING_EXIT;
               }
               break;
           } else if (mf == MazeFields.EMPTY || mf == MazeFields.FIRE) {
               m.setFire(py, px);
           } else {
               // Wall, exit
               break;
           }
           
        }
    }
    
    public void checkTimeout() {
        if ((System.currentTimeMillis() - lastActivity) > TIME_OUT) {
            if (state != RobotState.COMPLETED) {
                state = RobotState.TIMED_OUT;
            }
        }
    }
    
    // Clean maze without robots and fires
    public void checkExit(Maze m) {
        if (m.getMaze(posY, posX) == MazeFields.EXIT) {
            if (state == RobotState.LOOKING_EXIT) {
                state = RobotState.COMPLETED;
            }
        }
    }
    
    public void turnCompleted() {
        if (!isActive()) {
            return;
        }
        
        // Is there action to perform
        if (steps.size() <= stepsExecuted) {
            return;
        }
        stepsExecuted++;
    }
    
    // Hitted by another robot (killing)
    private void kill() {
        if (state == RobotState.NEED_KILL || state == RobotState.LOOKING_EXIT) {
            state =  RobotState.KILLED;
        }
    }

    public static RobotModel findRobotAtPosition(int px, int py, List<RobotModel> robots) {
        for(RobotModel rm : robots) {
            if (rm.getPosX() == px && rm.getPosY() == py) {
                return rm;
            }
        }
        return null;
    }

    public void addAction(Actions action) {
        steps.add(action);
        
    }

    @Override
    public String toString() {
        return "RobotModel [id=" + id + ", name=" + name + ", state=" + state + ", stepsExecuted=" + stepsExecuted
                + ", posX=" + posX + ", posY=" + posY + ", direction=" + direction + "]";
    }

    public RobotOrientation getDirection() {
        return direction;
    }

    public Actions[] getExecutedActions() {
        Actions[] result = new Actions[stepsExecuted];
        for(int i = 0; i < stepsExecuted; i++) {
            result[i] = steps.get(i);
        }
        return result;
    }

    public Actions[] getPlannedActions() {
        Actions[] result = new Actions[steps.size()-stepsExecuted];
        for(int i = stepsExecuted; i < steps.size(); i++) {
            result[i-stepsExecuted] = steps.get(i);
        }
        return result;
    }
    
    
}
